
/*
 * DEPEND.C	compute element dependencies and determine simulation order
 */


#include "depend.h"
#include "control.h"
#include "memory.h"
#include "symtbl.h"

#include <stdio.h>
#include <stdlib.h>


ELEMENT *unlink_first(ELEMENT **address_of_list)
{
  ELEMENT *element;

  element = *address_of_list;
  if (element == NULL) {
    /* list is empty */
    return NULL;
  }
  /* unlink and return first element of list */
  *address_of_list = element->next;
  return element;
}


ELEMENT *unlink_element(ELEMENT *element, ELEMENT **address_of_list)
{
  ELEMENT *elements;

  if (element == NULL) {
    /* this should never happen */
    error("illegal NULL argument for unlink_element");
  }
  elements = *address_of_list;
  if (elements == element) {
    /* element found, is first in list */
    *address_of_list = element->next;
    return element;
  }
  /* element is not first in list, go searching */
  while (elements != NULL) {
    if (elements->next == element) {
      /* element found, unlink and return */
      elements->next = element->next;
      return element;
    }
    /* element not found, try next one */
    elements = elements->next;
  }
  /* element is not in list */
  return NULL;
}


typedef struct dependency {
  int has_new_value;
  ELEMENT *depends_on;
} DEPENDENCY;


void set_dependency(DEPENDENCY *dependencies,
		    int signal_number,
		    ELEMENT *depends_on)
{
  static char errmsg[ERRMSG_SIZE];

  if (dependencies[signal_number].depends_on != NULL) {
    /* signal is already specified as output of another element */
    sprintf(errmsg,
	    "signal '%s' specified as output more than once",
	    nameof_global_signal(signal_number));
    error(errmsg);
  }
  dependencies[signal_number].depends_on = depends_on;
}


void set_dependencies(DEPENDENCY *dependencies,
		      int number_signals,
		      ELEMENT *elements)
{
  int signal_number;
  static char errmsg[ERRMSG_SIZE];

  /* reset dependencies */
  for (signal_number = 0; signal_number < number_signals; signal_number++) {
    dependencies[signal_number].has_new_value = 0;
    dependencies[signal_number].depends_on = NULL;
  }
  /* loop on elements */
  while (elements != NULL) {
    switch (elements->type) {
      case E_CONSTANT:
	signal_number = elements->u.constant.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_CONNECT:
	signal_number = elements->u.connect.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_ADD:
      case E_SUB:
      case E_MUL:
      case E_DIV:
	signal_number = elements->u.arith.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_DELAY:
	signal_number = elements->u.delay.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_PROBE:
	/* this element has no output signal */
	break;
      case E_STEP:
	signal_number = elements->u.step.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_SIN:
	signal_number = elements->u.sin.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_COS:
	signal_number = elements->u.cos.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_ATAN:
	signal_number = elements->u.atan.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      case E_SQRT:
	signal_number = elements->u.sqrt.out;
	set_dependency(dependencies, signal_number, elements);
	break;
      default:
	sprintf(errmsg,
		"cannot set dependency for element of unknown type %d",
		elements->type);
	error(errmsg);
	break;
    }
    elements = elements->next;
  }
  /* check that every signal is specified as output of any element */
  for (signal_number = 0; signal_number < number_signals; signal_number++) {
    if (dependencies[signal_number].depends_on == NULL) {
      sprintf(errmsg,
	      "signal '%s' never specified as output",
	      nameof_global_signal(signal_number));
      error(errmsg);
    }
  }
}


int input_signal_missing(ELEMENT *element,
			 DEPENDENCY *dependencies)
{
  int signal_number;
  static char errmsg[ERRMSG_SIZE];

  switch (element->type) {
    case E_CONSTANT:
	/* this element does not depend on any input */
	return -1;
    case E_CONNECT:
	signal_number = element->u.connect.in;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_ADD:
    case E_SUB:
    case E_MUL:
    case E_DIV:
	signal_number = element->u.arith.op1;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	signal_number = element->u.arith.op2;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_DELAY:
	/* ATTENTION: DELAY IS NOT DEPENDENT ON ITS INPUT SIGNAL! */
	return -1;
    case E_PROBE:
	signal_number = element->u.probe.in1;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	signal_number = element->u.probe.in2;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	signal_number = element->u.probe.in3;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	signal_number = element->u.probe.in4;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_STEP:
	signal_number = element->u.step.in;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_SIN:
	signal_number = element->u.sin.in;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_COS:
	signal_number = element->u.cos.in;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_ATAN:
	signal_number = element->u.atan.in_y;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	signal_number = element->u.atan.in_x;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    case E_SQRT:
	signal_number = element->u.sqrt.in;
	if (!dependencies[signal_number].has_new_value) {
	  return signal_number;
	}
	return -1;
    default:
	sprintf(errmsg,
		"cannot determine inputs for element of unknown type %d",
		element->type);
	error(errmsg);
	return -1;
  }
}


void flag_output_signals(ELEMENT *element,
			 DEPENDENCY *dependencies)
{
  int signal_number;
  static char errmsg[ERRMSG_SIZE];

  switch (element->type) {
    case E_CONSTANT:
	signal_number = element->u.constant.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_CONNECT:
	signal_number = element->u.connect.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_ADD:
    case E_SUB:
    case E_MUL:
    case E_DIV:
	signal_number = element->u.arith.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_DELAY:
	signal_number = element->u.delay.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_PROBE:
	/* this element has no output signal */
	break;
    case E_STEP:
	signal_number = element->u.step.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_SIN:
	signal_number = element->u.sin.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_COS:
	signal_number = element->u.cos.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_ATAN:
	signal_number = element->u.atan.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    case E_SQRT:
	signal_number = element->u.sqrt.out;
	dependencies[signal_number].has_new_value = 1;
	break;
    default:
	sprintf(errmsg,
		"cannot flag outputs for element of unknown type %d",
		element->type);
	error(errmsg);
	break;
  }
}


ELEMENT *ordered_elements(ELEMENT *unordered_elements, int number_signals)
{
  DEPENDENCY *dependencies;
  ELEMENT *ordered_elements;
  ELEMENT *stack;
  ELEMENT *element;
  int missing_input;
  static char errmsg[ERRMSG_SIZE];

  /* allocate dependency vector */
  dependencies = allocate(number_signals * sizeof(*dependencies), ARENA_3);
  /* for every signal, compute which element it depends on */
  set_dependencies(dependencies, number_signals, unordered_elements);
  /* now order the elements according to their dependencies */
  ordered_elements = NULL;
  stack = NULL;
  while (1) {
    /* stack is empty, get next element from unordered list */
    element = unlink_first(&unordered_elements);
    if (element == NULL) {
      /* stack is empty and no elements left: done */
      break;
    }
    /* push element to stack */
#ifdef CHECK_ORDERING
    printf("push\t");
    show_element(element);
#endif /* CHECK_ORDERING */
    element->next = stack;
    stack = element;
    while (stack != NULL) {
      /* check if any input signal for top element is not yet computed */
      missing_input = input_signal_missing(stack, dependencies);
      if (missing_input == -1) {
	/* none missing: can schedule computation of this element */
	/* pop element from stack */
	element = stack;
	stack = element->next;
#ifdef CHECK_ORDERING
	printf("pop\t");
	show_element(element);
#endif /* CHECK_ORDERING */
	/* flag output signals as computed */
	flag_output_signals(element, dependencies);
	/* link to ordered list */
	element->next = ordered_elements;
	ordered_elements = element;
      } else {
	/* signal missing: push element which output signal is needed */
	element = unlink_element(dependencies[missing_input].depends_on,
				 &unordered_elements);
	if (element == NULL) {
	  /* element is no longer in element list, i.e. already on stack */
	  /* the only chance for this is a cycle in the dependency graph */
	  sprintf(errmsg,
		  "signal '%s', cycle in dependency graph detected",
		  nameof_global_signal(missing_input));
	  error(errmsg);
	}
#ifdef CHECK_ORDERING
	printf("push\t");
	show_element(element);
#endif /* CHECK_ORDERING */
	element->next = stack;
	stack = element;
      }
    }
  }
  /* reverse ordered list of elements */
  stack = ordered_elements;
  ordered_elements = NULL;
  while (stack != NULL) {
    element = stack;
    stack = element->next;
    element->next = ordered_elements;
    ordered_elements = element;
  }
  /* deallocate dependency vector */
  deallocate_arena(ARENA_3);
  /* return ordered list of elements */
  return ordered_elements;
}
