
%{

/*
 * DSPLAN.L	DSP language lexical analyzer
 */


#include "control.h"
#include "memory.h"
#include "tree.h"
#include "ytab.h"

#include <string.h>

#define INPUT_STACK_SIZE	50

int yylook(void);
int yywrap(void);
int yyback(int *p, int m);

%}


WS		(\ |\t|\n)+
COM		\/\/.*
L		[A-Za-z_]
D		[0-9]
ID		{L}({L}|{D})*
NUM		{D}+
INT		(\+|\-)?{NUM}
FLO		(\+|\-)?{NUM}\.{NUM}(e(\+|\-)?{NUM})?
STR		\"[^\"\n]*(\"|\n)


%%


{WS}		{
		  /* nothing */
		}

{COM}		{
		  /* nothing */
		}

\#include	{
		  return INCLUDE;
		}

\#simulate	{
		  return SIMULATE;
		}

macro		{
		  return MACRO;
		}

signal		{
		  return SIGNAL;
		}

{ID}		{
		  yylval.sval = allocate(strlen(yytext) + 1, ARENA_1);
		  strcpy(yylval.sval, yytext);
		  return IDENT;
		}

{INT}		{
		  sscanf(yytext, "%d", &yylval.ival);
		  return INUM;
		}

{FLO}		{
		  sscanf(yytext, "%le", &yylval.fval);
		  return FNUM;
		}

{STR}		{
		  static char errmsg[ERRMSG_SIZE];
		  if (yytext[yyleng - 1] != '\"') {
		    sprintf(errmsg,
			    "file '%s', line %d, string exceeds line",
			    yyfilename, yylineno);
		    error(errmsg);
		  }
		  yytext[yyleng - 1] = '\0';
		  yylval.sval = allocate(strlen(yytext + 1) + 1, ARENA_1);
		  strcpy(yylval.sval, yytext + 1);
		  return STRING;
		}

\(		{
		  return LPAR;
		}

\)		{
		  return RPAR;
		}

\{		{
		  return LCURL;
		}

\}		{
		  return RCURL;
		}

\,		{
		  return COMMA;
		}

\;		{
		  return SEMI;
		}

.		{
		  static char errmsg[ERRMSG_SIZE];
		  sprintf(errmsg,
			  "file '%s', line %d, illegal character '%c'",
			  yyfilename, yylineno, yytext[0]);
		  error(errmsg);
		}


%%


/**************************************************************/


typedef struct select_record {
  char *name;
  struct select_record *next;
} SELECT_RECORD;

static SELECT_RECORD *selected_files = NULL;


int file_has_been_selected_before(char *filename)
{
  SELECT_RECORD *p;

  p = selected_files;
  while (p != NULL) {
    if (strcmp(p->name, filename) == 0) {
      return 1;
    }
    p = p->next;
  }
  return 0;
}


void add_to_selected_files(char *filename)
{
  SELECT_RECORD *p;

  p = allocate(sizeof(*p), ARENA_1);
  p->name = filename;
  p->next = selected_files;
  selected_files = p;
}


void reset_selected_files(void)
{
  selected_files = NULL;
}


/**************************************************************/


FILE *open_source(char *filename)
{
  static char lib_path[200];
  FILE *file;
  static char errmsg[ERRMSG_SIZE];

  /* try the current directory first */
  file = fopen(filename, "rt");
  if (file != NULL) {
    return file;
  }
  /* then try the standard library path */
  strcpy(lib_path, lib_dir);
  strcat(lib_path, filename);
  file = fopen(lib_path, "rt");
  if (file != NULL) {
    return file;
  }
  /* not found */
  sprintf(errmsg,
	  "cannot open description file '%s'",
	  filename);
  error(errmsg);
  /* never reached */
  return NULL;
}


/**************************************************************/


static struct {
  char *name;
  int line;
  YY_BUFFER_STATE state;
} input_stack[INPUT_STACK_SIZE];

static int input_sp = 0;

char *yyfilename = "???";


int yywrap(void)
{
  /* the following condition should never be true */
  if (input_sp == 0) {
    reset_selected_files();
    return 1;
  }
  /* close current input file */
  if (yyin != stdin) {
    fclose(yyin);
  }
  /* pop previous input file from stack */
  input_sp--;
  yyfilename = input_stack[input_sp].name;
  yylineno = input_stack[input_sp].line;
  yy_delete_buffer(YY_CURRENT_BUFFER);
  yy_switch_to_buffer(input_stack[input_sp].state);
  /* return 1 if no more input */
  if (input_sp == 0) {
    reset_selected_files();
    return 1;
  }
  return 0;
}


void yyselect(char *filename)
{
  /* do not select a file which has already been selected before */
  if (file_has_been_selected_before(filename)) {
    return;
  }
  add_to_selected_files(filename);
  /* push current input file to stack */
  if (input_sp == INPUT_STACK_SIZE) {
    error("too many nested inclusions");
  }
  input_stack[input_sp].name = yyfilename;
  input_stack[input_sp].line = yylineno;
  input_stack[input_sp].state = YY_CURRENT_BUFFER;
  input_sp++;
  /* open new input file */
  yyfilename = filename;
  yyin = open_source(yyfilename);
  yylineno = 1;
  yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
}


void yycleanup(void)
{
  while (!yywrap()) ;
}


/**************************************************************/


void show_token(int token)
{
  static char errmsg[ERRMSG_SIZE];

  switch (token) {
    case INCLUDE:
	printf("INCLUDE");
	break;
    case SIMULATE:
	printf("SIMULATE");
	break;
    case MACRO:
	printf("MACRO");
	break;
    case SIGNAL:
	printf("SIGNAL");
	break;
    case IDENT:
	printf("IDENT\t\t%s", yylval.sval);
	break;
    case INUM:
	printf("INUM\t\t%d", yylval.ival);
	break;
    case FNUM:
	printf("FNUM\t\t%e", yylval.fval);
	break;
    case STRING:
	printf("STRING\t\t\"%s\"", yylval.sval);
	break;
    case LPAR:
	printf("LPAR");
	break;
    case RPAR:
	printf("RPAR");
	break;
    case LCURL:
	printf("LCURL");
	break;
    case RCURL:
	printf("RCURL");
	break;
    case COMMA:
	printf("COMMA");
	break;
    case SEMI:
	printf("SEMI");
	break;
    default:
	sprintf(errmsg,
		"file '%s', line %d, cannot show unknown token %d",
		yyfilename, yylineno, token);
	error(errmsg);
	break;
  }
  printf("\n");
}
