
/*
 * ELEMENT.C	generation of DSP simulation elements
 */


#include "element.h"
#include "control.h"
#include "memory.h"
#include "symtbl.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static ELEMENT *total_elements;


static int stack[STACKSIZE];
static int sp, bp;


static void push(int val)
{
  if (sp >= STACKSIZE - 1) {
    error("macro expansion stack overflow");
  }
  stack[++sp] = val;
}


static int pop(void)
{
  if (sp < 0) {
    error("macro expansion stack underflow");
  }
  return stack[sp--];
}


void new_constant(double value, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_CONSTANT;
  element->u.constant.value = value;
  element->u.constant.out = out;
  element->next = total_elements;
  total_elements = element;
}


int resolve(NODE *argument)
{
  SYMBOL *signal;
  static int generated_constants = 0;
  static char name[200];
  char *constant_name;
  static char errmsg[ERRMSG_SIZE];

  switch (argument->type) {
    case N_ARG_GLOBSIG:
	return argument->u.arg_globsig.signum;
    case N_ARG_PARAM:
	return stack[bp + 2 + argument->u.arg_param.parmnum];
    case N_ARG_LOCSIG:
	return stack[bp + 1] + argument->u.arg_locsig.signum;
    case N_ARG_INTVAL:
	sprintf(name, "_%d", generated_constants++);
	constant_name = allocate(strlen(name) + 1, ARENA_1);
	strcpy(constant_name, name);
	signal = enter_global_signal(constant_name);
	new_constant(argument->u.arg_intval.value, signal->info.number);
	return signal->info.number;
    case N_ARG_FLOVAL:
	sprintf(name, "_%d", generated_constants++);
	constant_name = allocate(strlen(name) + 1, ARENA_1);
	strcpy(constant_name, name);
	signal = enter_global_signal(constant_name);
	new_constant(argument->u.arg_floval.value, signal->info.number);
	return signal->info.number;
    default:
	sprintf(errmsg,
		"cannot resolve unknown argument type %d",
		argument->type);
	error(errmsg);
	return 0;
  }
}


void new_connect(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_CONNECT;
  element->u.connect.in = in;
  element->u.connect.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_arith(int type, int op1, int op2, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = type;
  element->u.arith.op1 = op1;
  element->u.arith.op2 = op2;
  element->u.arith.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_delay(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_DELAY;
  element->u.delay.in = in;
  element->u.delay.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_probe(int in1, int in2, int in3, int in4)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_PROBE;
  element->u.probe.in1 = in1;
  element->u.probe.in2 = in2;
  element->u.probe.in3 = in3;
  element->u.probe.in4 = in4;
  element->next = total_elements;
  total_elements = element;
}


void new_step(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_STEP;
  element->u.step.in = in;
  element->u.step.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_sin(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_SIN;
  element->u.sin.in = in;
  element->u.sin.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_cos(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_COS;
  element->u.cos.in = in;
  element->u.cos.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_atan(int in_y, int in_x, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_ATAN;
  element->u.atan.in_y = in_y;
  element->u.atan.in_x = in_x;
  element->u.atan.out = out;
  element->next = total_elements;
  total_elements = element;
}


void new_sqrt(int in, int out)
{
  ELEMENT *element;

  element = allocate(sizeof(*element), ARENA_2);
  element->type = E_SQRT;
  element->u.sqrt.in = in;
  element->u.sqrt.out = out;
  element->next = total_elements;
  total_elements = element;
}


void gen_builtin(NODE *builtin, NODE *args)
{
  static char errmsg[ERRMSG_SIZE];

  switch (builtin->u.builtin.type) {
    case B_CONNECT:
	new_connect(resolve(args),
		    resolve(args->next));
	break;
    case B_ADD:
	new_arith(E_ADD,
		  resolve(args),
		  resolve(args->next),
		  resolve(args->next->next));
	break;
    case B_SUB:
	new_arith(E_SUB,
		  resolve(args),
		  resolve(args->next),
		  resolve(args->next->next));
	break;
    case B_MUL:
	new_arith(E_MUL,
		  resolve(args),
		  resolve(args->next),
		  resolve(args->next->next));
	break;
    case B_DIV:
	new_arith(E_DIV,
		  resolve(args),
		  resolve(args->next),
		  resolve(args->next->next));
	break;
    case B_DELAY:
	new_delay(resolve(args),
		  resolve(args->next));
	break;
    case B_PROBE:
	new_probe(resolve(args),
	          resolve(args->next),
	          resolve(args->next->next),
	          resolve(args->next->next->next));
	break;
    case B_STEP:
	new_step(resolve(args),
		 resolve(args->next));
	break;
    case B_SIN:
	new_sin(resolve(args),
		resolve(args->next));
	break;
    case B_COS:
	new_cos(resolve(args),
		resolve(args->next));
	break;
    case B_ATAN:
	new_atan(resolve(args),
		 resolve(args->next),
		 resolve(args->next->next));
	break;
    case B_SQRT:
	new_sqrt(resolve(args),
		 resolve(args->next));
	break;
    default:
	sprintf(errmsg,
		"cannot generate unknown builtin type %d",
		builtin->u.builtin.type);
	error(errmsg);
	break;
  }
}


void gen_elements_for_nodelist(NODE *nodelist);


void gen_macro(NODE *macro, NODE *args)
{
  int numparms;
  int numlocals;
  NODE *calls;
  int new_bp;
  static int macro_expansions = 0;
  static char name[200];
  char *local_name;
  int i;

  numparms = macro->u.macro.numparms;
  numlocals = macro->u.macro.numlocals;
  calls = macro->u.macro.calls;
  push(bp);
  new_bp = sp;
  push(0);	/* dummy, will become first_local_signal of new frame */
  for (i = 0; i < numparms; i++) {
    push(resolve(args));
    args = args->next;
  }
  bp = new_bp;
  stack[bp + 1] = get_number_global_signals();
  for (i = 0; i < numlocals; i++) {
    sprintf(name,
	    "%s.%d.local.%d",
	    macro->u.macro.name,
	    macro_expansions++,
	    i);
    local_name = allocate(strlen(name) + 1, ARENA_1);
    strcpy(local_name, name);
    enter_global_signal(local_name);
  }
  gen_elements_for_nodelist(calls);
  sp = bp;
  bp = pop();
}


void gen_elements_for_node(NODE *node)
{
  static char errmsg[ERRMSG_SIZE];

  if (node->type != N_CALL) {
    sprintf(errmsg,
	    "cannot generate elements for node type %d",
	    node->type);
    error(errmsg);
    return;
  }
  if (node->u.call.callee->type == N_BUILTIN) {
    gen_builtin(node->u.call.callee, node->u.call.args);
    return;
  }
  if (node->u.call.callee->type == N_MACRO) {
    gen_macro(node->u.call.callee, node->u.call.args);
    return;
  }
  sprintf(errmsg,
	  "illegal callee type %d",
	  node->u.call.callee->type);
  error(errmsg);
}


void gen_elements_for_nodelist(NODE *nodelist)
{
  while (nodelist != NULL) {
    gen_elements_for_node(nodelist);
    nodelist = nodelist->next;
  }
}


ELEMENT *gen_elements(NODE *nodelist)
{
  total_elements = NULL;
  sp = -1;
  bp = -1;
  gen_elements_for_nodelist(nodelist);
  return total_elements;
}


void show_element(ELEMENT *element)
{
  static char errmsg[ERRMSG_SIZE];

  switch (element->type) {
    case E_CONSTANT:
	printf("CONSTANT ( %e, %d )",
	       element->u.constant.value,
	       element->u.constant.out);
	break;
    case E_CONNECT:
	printf("CONNECT ( %d, %d )",
	       element->u.connect.in,
	       element->u.connect.out);
	break;
    case E_ADD:
	printf("ADD ( %d, %d, %d )",
	       element->u.arith.op1,
	       element->u.arith.op2,
	       element->u.arith.out);
	break;
    case E_SUB:
	printf("SUB ( %d, %d, %d )",
	       element->u.arith.op1,
	       element->u.arith.op2,
	       element->u.arith.out);
	break;
    case E_MUL:
	printf("MUL ( %d, %d, %d )",
	       element->u.arith.op1,
	       element->u.arith.op2,
	       element->u.arith.out);
	break;
    case E_DIV:
	printf("DIV ( %d, %d, %d )",
	       element->u.arith.op1,
	       element->u.arith.op2,
	       element->u.arith.out);
	break;
    case E_DELAY:
	printf("DELAY ( %d, %d )",
	       element->u.delay.in,
	       element->u.delay.out);
	break;
    case E_PROBE:
	printf("PROBE ( %d, %d, %d, %d )",
	       element->u.probe.in1,
	       element->u.probe.in2,
	       element->u.probe.in3,
	       element->u.probe.in4);
	break;
    case E_STEP:
	printf("STEP ( %d, %d )",
	       element->u.step.in,
	       element->u.step.out);
	break;
    case E_SIN:
	printf("SIN ( %d, %d )",
	       element->u.sin.in,
	       element->u.sin.out);
	break;
    case E_COS:
	printf("COS ( %d, %d )",
	       element->u.cos.in,
	       element->u.cos.out);
	break;
    case E_ATAN:
	printf("ATAN ( %d, %d, %d )",
	       element->u.atan.in_y,
	       element->u.atan.in_x,
	       element->u.atan.out);
	break;
    case E_SQRT:
	printf("SQRT ( %d, %d )",
	       element->u.sqrt.in,
	       element->u.sqrt.out);
	break;
    default:
	sprintf(errmsg,
		"cannot show unknown element type %d",
		element->type);
	error(errmsg);
	break;
  }
  printf("\n");
}


void show_elements(ELEMENT *elements)
{
  while (elements != NULL) {
    show_element(elements);
    elements = elements->next;
  }
}
