
/*
 * SIMULATE.C	DSP simulation kernel
 */


#include "simulate.h"
#include "control.h"
#include "memory.h"
#include "symtbl.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#define EPS	1e-10		/* this means zero, numerically */
#define PI	M_PI


ELEMENT *init_delays(ELEMENT *elements)
{
  ELEMENT *delays;

  delays = NULL;
  while (elements != NULL) {
    if (elements->type == E_DELAY) {
      elements->u.delay.state = 0.0;
      elements->u.delay.next = delays;
      delays = elements;
    }
    elements = elements->next;
  }
  return delays;
}


void write_ctrl(ELEMENT *elements, int number_cycles)
{
  int number_signals;
  ELEMENT *p;
  char *signal_name;

  fprintf(ctrl_out, "%d\n", number_cycles);
  number_signals = 0;
  p = elements;
  while (p != NULL) {
    if (p->type == E_PROBE) {
      number_signals += 4;
    }
    p = p->next;
  }
  fprintf(ctrl_out, "%d\n", number_signals);
  p = elements;
  while (p != NULL) {
    if (p->type == E_PROBE) {
      signal_name = nameof_global_signal(p->u.probe.in1);
      fprintf(ctrl_out, "%s\n", signal_name);
      signal_name = nameof_global_signal(p->u.probe.in2);
      fprintf(ctrl_out, "%s\n", signal_name);
      signal_name = nameof_global_signal(p->u.probe.in3);
      fprintf(ctrl_out, "%s\n", signal_name);
      signal_name = nameof_global_signal(p->u.probe.in4);
      fprintf(ctrl_out, "%s\n", signal_name);
    }
    p = p->next;
  }
}


void simulate(ELEMENT *elements, int number_signals, int number_cycles)
{
  ELEMENT *delays;
  double *signals;
  int cycle;
  ELEMENT *element;
  static char errmsg[ERRMSG_SIZE];

  /* reset state variables and link all delays into a list */
  delays = init_delays(elements);
  /* allocate signal vector */
  signals = allocate(number_signals * sizeof(*signals), ARENA_4);
  /* write control file */
  write_ctrl(elements, number_cycles);
  /* do simulation */
  for (cycle = 0; cycle < number_cycles; cycle++) {
    for (element = elements; element != NULL; element = element->next) {
      switch (element->type) {
	case E_CONSTANT:
		/* supply a constant value */
		signals[element->u.constant.out] =
		  element->u.constant.value;
		break;
	case E_CONNECT:
		/* connect a signal to another one */
		signals[element->u.connect.out] =
		  signals[element->u.connect.in];
		break;
	case E_ADD:
		/* add */
		signals[element->u.arith.out] =
		  signals[element->u.arith.op1] +
		  signals[element->u.arith.op2];
		break;
	case E_SUB:
		/* subtract */
		signals[element->u.arith.out] =
		  signals[element->u.arith.op1] -
		  signals[element->u.arith.op2];
		break;
	case E_MUL:
		/* multiply */
		signals[element->u.arith.out] =
		  signals[element->u.arith.op1] *
		  signals[element->u.arith.op2];
		break;
	case E_DIV:
		/* divide */
		if (fabs(signals[element->u.arith.op2]) <= EPS) {
		  sprintf(errmsg,
			  "signal '%s', division by zero",
			  nameof_global_signal(element->u.arith.op2));
		  error(errmsg);
		}
		signals[element->u.arith.out] =
		  signals[element->u.arith.op1] /
		  signals[element->u.arith.op2];
		break;
	case E_DELAY:
		/* delay for one step, part 1: state -> output */
		signals[element->u.delay.out] =
		  element->u.delay.state;
		break;
	case E_PROBE:
		/* record value */
		fwrite(&signals[element->u.probe.in1],
		       sizeof(double), 1, data_out);
		fwrite(&signals[element->u.probe.in2],
		       sizeof(double), 1, data_out);
		fwrite(&signals[element->u.probe.in3],
		       sizeof(double), 1, data_out);
		fwrite(&signals[element->u.probe.in4],
		       sizeof(double), 1, data_out);
		break;
	case E_STEP:
		/* step function */
		if (signals[element->u.step.in] <= EPS) {
		  signals[element->u.step.out] = 0.0;
		} else {
		  signals[element->u.step.out] = 1.0;
		}
		break;
	case E_SIN:
		/* sine */
		signals[element->u.sin.out] =
		  sin(2 * PI * signals[element->u.sin.in]);
		break;
	case E_COS:
		/* cosine */
		signals[element->u.cos.out] =
		  cos(2 * PI * signals[element->u.cos.in]);
		break;
	case E_ATAN:
		/* arc-tangent */
		signals[element->u.atan.out] =
		  atan2(signals[element->u.atan.in_y],
			signals[element->u.atan.in_x]) / (2 * PI);
		break;
	case E_SQRT:
		/* square root  */
		signals[element->u.sqrt.out] =
		  sqrt(signals[element->u.sqrt.in]);
		break;
	default:
		sprintf(errmsg,
			"cannot simulate unknown element type %d",
			element->type);
		error(errmsg);
		break;
      }
    }
    /* handle delays, part2: input -> state */
    for (element = delays; element != NULL; element = element->u.delay.next) {
      element->u.delay.state = signals[element->u.delay.in];
    }
  }
  /* deallocate signal vector */
  deallocate_arena(ARENA_4);
}
