
/*
 * SYMTBL.C	symbol table management
 */


#include "symtbl.h"
#include "control.h"
#include "memory.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/**************************************************************/


static SYMBOL *lookup(SYMBOL *symbol, char *name)
{
  while (symbol != NULL) {
    if (strcmp(symbol->name, name) == 0) {
      return symbol;
    }
    symbol = symbol->next;
  }
  return NULL;
}


static SYMBOL *enter(SYMBOL **anchor, char *name)
{
  SYMBOL *symbol;
  static char errmsg[ERRMSG_SIZE];

  symbol = lookup(*anchor, name);
  if (symbol != NULL) {
    sprintf(errmsg,
	    "file '%s', line %d, multiple declarations for symbol '%s'",
	    yyfilename, yylineno, name);
    error(errmsg);
  }
  symbol = allocate(sizeof(*symbol), ARENA_1);
  symbol->name = name;
  symbol->next = *anchor;
  *anchor = symbol;
  return symbol;
}


/**************************************************************/


static struct {
  char *name;
  int type;
  int numparms;
} builtin_tbl[] = {
  { "connect",    B_CONNECT,    2 },
  { "add",        B_ADD,        3 },
  { "sub",        B_SUB,        3 },
  { "mul",        B_MUL,        3 },
  { "div",        B_DIV,        3 },
  { "delay",      B_DELAY,      2 },
  { "probe",      B_PROBE,      4 },
  { "step",       B_STEP,       2 },
  { "sin",        B_SIN,        2 },
  { "cos",        B_COS,        2 },
  { "atan",       B_ATAN,       3 },
  { "sqrt",       B_SQRT,       2 }
};

static SYMBOL *builtins;


void init_builtins(void)
{
  int i;
  SYMBOL *symbol;

  builtins = NULL;
  for (i = 0; i < sizeof(builtin_tbl) / sizeof(builtin_tbl[0]); i++) {
    symbol = enter(&builtins, builtin_tbl[i].name);
    symbol->info.node =
      new_builtin(builtin_tbl[i].name,
		  builtin_tbl[i].type,
		  builtin_tbl[i].numparms);
  }
}


SYMBOL *lookup_builtin(char *name)
{
  return lookup(builtins, name);
}


/**************************************************************/


static SYMBOL *macros;


void init_macros(void)
{
  macros = NULL;
}


SYMBOL *lookup_macro(char *name)
{
  return lookup(macros, name);
}


SYMBOL *enter_macro(char *name, NODE *macro)
{
  SYMBOL *symbol;

  symbol = enter(&macros, name);
  symbol->info.node = macro;
  return symbol;
}


void show_macros(void)
{
  SYMBOL *symbol;

  symbol = macros;
  while (symbol != NULL) {
    show_node(symbol->info.node, 1);
    symbol = symbol->next;
  }
}


/**************************************************************/


static SYMBOL *global_signals;
static int number_global_signals;


void init_global_signals(void)
{
  global_signals = NULL;
  number_global_signals = 0;
}


SYMBOL *lookup_global_signal(char *name)
{
  return lookup(global_signals, name);
}


SYMBOL *enter_global_signal(char *name)
{
  SYMBOL *symbol;

  symbol = enter(&global_signals, name);
  symbol->info.number = number_global_signals++;
  return symbol;
}


char *nameof_global_signal(int signal_number)
{
  SYMBOL *symbol;

  symbol = global_signals;
  while (symbol != NULL) {
    if (symbol->info.number == signal_number) {
      return symbol->name;
    }
    symbol = symbol->next;
  }
  return "global signal not found";
}


int get_number_global_signals(void)
{
  return number_global_signals;
}


/**************************************************************/


static SYMBOL *params;
static int number_params;


void init_params(void)
{
  params = NULL;
  number_params = 0;
}


SYMBOL *lookup_param(char *name)
{
  return lookup(params, name);
}


SYMBOL *enter_param(char *name)
{
  SYMBOL *symbol;

  symbol = enter(&params, name);
  symbol->info.number = number_params++;
  return symbol;
}


int get_number_params(void)
{
  return number_params;
}


/**************************************************************/


static SYMBOL *local_signals;
static int number_local_signals;


void init_local_signals(void)
{
  local_signals = NULL;
  number_local_signals = 0;
}


SYMBOL *lookup_local_signal(char *name)
{
  return lookup(local_signals, name);
}


SYMBOL *enter_local_signal(char *name)
{
  SYMBOL *symbol;

  symbol = enter(&local_signals, name);
  symbol->info.number = number_local_signals++;
  return symbol;
}


int get_number_local_signals(void)
{
  return number_local_signals;
}
